/* The copyright in this software is being made available under the BSD
 * License, included below. This software may be subject to other third party
 * and contributor rights, including patent rights, and no such rights are
 * granted under this license.
 *
 * Copyright (c) 2010-2020, ISO/IEC
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 *
 *  * Redistributions of source code must retain the above copyright notice,
 *    this list of conditions and the following disclaimer.
 *  * Redistributions in binary form must reproduce the above copyright notice,
 *    this list of conditions and the following disclaimer in the documentation
 *    and/or other materials provided with the distribution.
 *  * Neither the name of the ISO/IEC nor the names of its contributors may
 *    be used to endorse or promote products derived from this software without
 *    specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 * AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
 * THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <TMIV/Renderer/RecoverPrunedViews.h>

#include <TMIV/MivBitstream/DepthOccupancyTransform.h>

#include <algorithm>
#include <iostream>

using namespace std;
using namespace TMIV::Common;
using namespace TMIV::MivBitstream;

namespace TMIV::Renderer {
// NOTE(BK): This new implementation relies on the block to patch map. There is no assumption on
// patch ordering anymore.
auto recoverPrunedViewAndMask(const AccessUnit &frame)
    -> pair<vector<Texture444Depth10Frame>, MaskList> {
  // Initialization
  auto prunedView = vector<Texture444Depth10Frame>{};
  auto prunedMasks = MaskList{};

  const auto &viewParamsList = frame.atlas.front().viewParamsList;

  for (const auto &viewParams : viewParamsList) {
    const auto size = viewParams.ci.projectionPlaneSize();
    prunedView.emplace_back(Texture444Frame{size.x(), size.y()}, Depth10Frame{size.x(), size.y()});
    prunedView.back().first.fillNeutral();
    prunedMasks.emplace_back(size.x(), size.y());
    prunedMasks.back().fillZero();
  }

  // For each pixel in each atlas
  for (const auto &atlas : frame.atlas) {
    for (int i = 0; i < atlas.asps.asps_frame_height(); ++i) {
      for (int j = 0; j < atlas.asps.asps_frame_width(); ++j) {
        // Fetch patch ID
        const auto patchId = atlas.patchId(i, j);
        if (patchId == unusedPatchId) {
          continue;
        }

        // Index patch and view parameters
        const auto &patchParams = atlas.patchParamsList[patchId];
        const auto viewId = patchParams.pduViewId();
        const auto &viewParams = atlas.viewParamsList[viewId];

        // Test for occupancy
        const auto occupancyTransform = OccupancyTransform{viewParams, patchParams};
        if (!occupancyTransform.occupant(atlas.geoFrame.getPlane(0)(i, j))) {
          continue;
        }

        // Map to view position
        const auto viewPos = patchParams.atlasToView({j, i});
        const auto x = viewPos.x();
        const auto y = viewPos.y();

        // Copy geometry
        prunedView[viewId].second.getPlane(0)(y, x) = atlas.geoFrame.getPlane(0)(i, j);

        // Copy attributes
        for (int d = 0; d < 3; ++d) {
          prunedView[viewId].first.getPlane(d)(y, x) = atlas.attrFrame.getPlane(d)(i, j);
        }

        // Set mask
        prunedMasks[viewId].getPlane(0)(y, x) = UINT8_MAX;
      }
    }
  }

  return pair{prunedView, prunedMasks};
}
} // namespace TMIV::Renderer
